from dagster import ModeDefinition, execute_solid, solid
from dagster_aws.ecr import fake_ecr_public_resource


def test_ecr_public_get_login_password():
    @solid(required_resource_keys={"ecr_public"})
    def ecr_public_solid(context):
        return context.resources.ecr_public.get_login_password()

    result = execute_solid(
        ecr_public_solid,
        mode_def=ModeDefinition(resource_defs={"ecr_public": fake_ecr_public_resource}),
    )

    assert result.output_value() == "token"



from queue import Queue
from threading import Thread

from dagster import DagsterEvent
from dagster.core.events import EngineEventData
from dagster_aws.emr.emr_step_main import DONE, event_writing_loop


def make_event(event_id):
    return DagsterEvent(
        event_type_value="ENGINE_EVENT",
        pipeline_name="some_pipeline",
        event_specific_data=EngineEventData(),
        message=str(event_id),
    )


EVENTS = [make_event(i) for i in range(3)]


def start_event_writing_thread(events_queue):
    """Returns the thread and a queue with an entry for each list of events written, so that the
    caller can inspect what was written.
    """
    written_events = Queue()

    def put_events_in_queue(events):
        written_events.put(events)

    event_writing_thread = Thread(
        target=event_writing_loop,
        kwargs=dict(
            events_queue=events_queue,
            put_events_fn=put_events_in_queue,
        ),
    )
    event_writing_thread.start()
    return event_writing_thread, written_events


def test_done_ends_event_writing_thread():
    events_queue = Queue()
    try:
        event_writing_thread, _ = start_event_writing_thread(events_queue)
    finally:
        events_queue.put(DONE)
    event_writing_thread.join(timeout=2)
    assert not event_writing_thread.is_alive()


def test_write_events():
    events_queue = Queue()
    try:
        event_writing_thread, written_events = start_event_writing_thread(events_queue)
        events_queue.put(EVENTS[0])
    finally:
        events_queue.put(DONE)
    event_writing_thread.join(timeout=2)
    assert not event_writing_thread.is_alive()
    assert written_events.get(timeout=2) == [EVENTS[0]]


def test_rewrite_earlier_events():
    events_queue = Queue()
    try:
        event_writing_thread, written_events = start_event_writing_thread(events_queue)
        events_queue.put(EVENTS[0])
        assert written_events.get(timeout=2) == EVENTS[0:1]

        events_queue.put(EVENTS[1])
        assert written_events.get(timeout=2) == EVENTS[0:2]
    finally:
        events_queue.put(DONE)
    event_writing_thread.join(timeout=2)
    assert not event_writing_thread.is_alive()



def test_termination(instance, workspace, run):
    assert not instance.run_launcher.can_terminate(run.run_id)

    instance.launch_run(run.run_id, workspace)

    assert instance.run_launcher.can_terminate(run.run_id)
    assert instance.run_launcher.terminate(run.run_id)
    assert not instance.run_launcher.can_terminate(run.run_id)
    assert not instance.run_launcher.terminate(run.run_id)


def test_missing_run(instance, workspace, run, monkeypatch):
    instance.launch_run(run.run_id, workspace)

    def missing_run(*_args, **_kwargs):
        return None

    original = instance.get_run_by_id

    monkeypatch.setattr(instance, "get_run_by_id", missing_run)
    assert not instance.run_launcher.can_terminate(run.run_id)

    monkeypatch.setattr(instance, "get_run_by_id", original)
    assert instance.run_launcher.can_terminate(run.run_id)

    monkeypatch.setattr(instance, "get_run_by_id", missing_run)
    assert not instance.run_launcher.terminate(run.run_id)

    monkeypatch.setattr(instance, "get_run_by_id", original)
    assert instance.run_launcher.terminate(run.run_id)


def test_missing_tag(instance, workspace, run):
    instance.launch_run(run.run_id, workspace)
    original = instance.get_run_by_id(run.run_id).tags

    instance.add_run_tags(run.run_id, {"ecs/task_arn": ""})
    assert not instance.run_launcher.can_terminate(run.run_id)

    instance.add_run_tags(run.run_id, original)
    instance.add_run_tags(run.run_id, {"ecs/cluster": ""})
    assert not instance.run_launcher.can_terminate(run.run_id)

    instance.add_run_tags(run.run_id, original)
    assert instance.run_launcher.can_terminate(run.run_id)

    instance.add_run_tags(run.run_id, {"ecs/task_arn": ""})
    assert not instance.run_launcher.terminate(run.run_id)

    instance.add_run_tags(run.run_id, original)
    instance.add_run_tags(run.run_id, {"ecs/cluster": ""})
    assert not instance.run_launcher.terminate(run.run_id)

    instance.add_run_tags(run.run_id, original)
    assert instance.run_launcher.terminate(run.run_id)


def test_eventual_consistency(instance, workspace, run, monkeypatch):
    instance.launch_run(run.run_id, workspace)

    def empty(*_args, **_kwargs):
        return {"tasks": []}

    original = instance.run_launcher.ecs.describe_tasks

    monkeypatch.setattr(instance.run_launcher.ecs, "describe_tasks", empty)
    assert not instance.run_launcher.can_terminate(run.run_id)

    monkeypatch.setattr(instance.run_launcher.ecs, "describe_tasks", original)
    assert instance.run_launcher.can_terminate(run.run_id)

    monkeypatch.setattr(instance.run_launcher.ecs, "describe_tasks", empty)
    assert not instance.run_launcher.terminate(run.run_id)

    monkeypatch.setattr(instance.run_launcher.ecs, "describe_tasks", original)
    assert instance.run_launcher.terminate(run.run_id)



# pylint: disable=redefined-outer-name
import boto3
import moto
import pytest

from .stubbed_ecs import StubbedEcs


@pytest.fixture
def region():
    return "us-east-1"


@pytest.fixture
def ecs(region):
    return StubbedEcs(boto3.client("ecs", region_name=region))


@pytest.fixture
def ec2(region):
    with moto.mock_ec2():
        yield boto3.resource("ec2", region_name=region)


@pytest.fixture
def vpc(ec2):
    return ec2.create_vpc(CidrBlock="10.0.0.0/16")


@pytest.fixture
def subnet(vpc):
    return vpc.create_subnet(CidrBlock="10.0.0.0/24")


@pytest.fixture
def security_group(vpc):
    return vpc.create_security_group(Description="test", GroupName="test")


@pytest.fixture
def network_interface(subnet, security_group):
    return subnet.create_network_interface(Groups=[security_group.id])



import io
import pickle

from dagster import Field, IOManager, StringSource, check, io_manager
from dagster.utils import PICKLE_PROTOCOL


class PickledObjectS3IOManager(IOManager):
    def __init__(
        self,
        s3_bucket,
        s3_session,
        s3_prefix=None,
    ):
        self.bucket = check.str_param(s3_bucket, "s3_bucket")
        self.s3_prefix = check.str_param(s3_prefix, "s3_prefix")
        self.s3 = s3_session
        self.s3.head_bucket(Bucket=self.bucket)

    def _get_path(self, context):
        return "/".join([self.s3_prefix, "storage", *context.get_output_identifier()])

    def _rm_object(self, key):
        check.str_param(key, "key")
        check.param_invariant(len(key) > 0, "key")

        def delete_for_results(store, results):
            store.s3.delete_objects(
                Bucket=store.bucket,
                Delete={"Objects": [{"Key": result["Key"]} for result in results["Contents"]]},
            )

        if self._has_object(key):
            results = self.s3.list_objects_v2(Bucket=self.bucket, Prefix=key)
            delete_for_results(self, results)

            continuation = results["IsTruncated"]
            while continuation:
                continuation_token = results["NextContinuationToken"]
                results = self.s3.list_objects_v2(
                    Bucket=self.bucket, Prefix=key, ContinuationToken=continuation_token
                )
                delete_for_results(self, results)
                continuation = results["IsTruncated"]

    def _has_object(self, key):
        check.str_param(key, "key")
        check.param_invariant(len(key) > 0, "key")

        key_count = self.s3.list_objects_v2(Bucket=self.bucket, Prefix=key)["KeyCount"]
        return bool(key_count > 0)

    def _uri_for_key(self, key):
        check.str_param(key, "key")
        return "s3://" + self.bucket + "/" + "{key}".format(key=key)

    def load_input(self, context):
        key = self._get_path(context.upstream_output)
        context.log.debug(f"Loading S3 object from: {self._uri_for_key(key)}")
        obj = pickle.loads(self.s3.get_object(Bucket=self.bucket, Key=key)["Body"].read())

        return obj

    def handle_output(self, context, obj):
        key = self._get_path(context)
        context.log.debug(f"Writing S3 object at: {self._uri_for_key(key)}")

        if self._has_object(key):
            context.log.warning(f"Removing existing S3 key: {key}")
            self._rm_object(key)

        pickled_obj = pickle.dumps(obj, PICKLE_PROTOCOL)
        pickled_obj_bytes = io.BytesIO(pickled_obj)
        self.s3.upload_fileobj(pickled_obj_bytes, self.bucket, key)


@io_manager(
    config_schema={
        "s3_bucket": Field(StringSource),
        "s3_prefix": Field(StringSource, is_required=False, default_value="dagster"),
    },
    required_resource_keys={"s3"},
)
def s3_pickle_io_manager(init_context):
    """Persistent IO manager using S3 for storage.

    Serializes objects via pickling. Suitable for objects storage for distributed executors, so long
    as each execution node has network connectivity and credentials for S3 and the backing bucket.

    Attach this resource definition to a :py:class:`~dagster.ModeDefinition`
    in order to make it available to your pipeline:

    .. code-block:: python

        pipeline_def = PipelineDefinition(
            mode_defs=[
                ModeDefinition(
                    resource_defs={'io_manager': s3_pickle_io_manager, "s3": s3_resource, ...},
                ), ...
            ], ...
        )

    You may configure this storage as follows:

    .. code-block:: YAML

        resources:
            io_manager:
                config:
                    s3_bucket: my-cool-bucket
                    s3_prefix: good/prefix-for-files-
    """
    s3_session = init_context.resources.s3
    s3_bucket = init_context.resource_config["s3_bucket"]
    s3_prefix = init_context.resource_config.get("s3_prefix")  # s3_prefix is optional
    pickled_io_manager = PickledObjectS3IOManager(s3_bucket, s3_session, s3_prefix=s3_prefix)
    return pickled_io_manager



# pylint: disable=protected-access
import pytest
from botocore.exceptions import ClientError, ParamValidationError


def test_describe_task_definition(ecs):
    with pytest.raises(ClientError):
        # The task definition doesn't exist
        ecs.describe_task_definition(taskDefinition="dagster")

    dagster1 = ecs.register_task_definition(
        family="dagster",
        containerDefinitions=[{"image": "hello_world:latest"}],
        networkMode="bridge",
    )
    dagster2 = ecs.register_task_definition(
        family="dagster",
        containerDefinitions=[{"image": "hello_world:latest"}],
    )

    # It gets the latest revision
    assert ecs.describe_task_definition(taskDefinition="dagster") == dagster2
    # It gets the specific revision
    assert ecs.describe_task_definition(taskDefinition="dagster:1") == dagster1
    assert ecs.describe_task_definition(taskDefinition="dagster:2") == dagster2

    # It also works with ARNs
    dagster1_arn = dagster1["taskDefinition"]["taskDefinitionArn"]
    dagster2_arn = dagster2["taskDefinition"]["taskDefinitionArn"]
    assert ecs.describe_task_definition(taskDefinition=dagster1_arn) == dagster1
    assert ecs.describe_task_definition(taskDefinition=dagster2_arn) == dagster2

    with pytest.raises(ClientError):
        # The revision doesn't exist
        ecs.describe_task_definition(taskDefinition="dagster:3")


def test_describe_tasks(ecs):
    assert not ecs.describe_tasks(tasks=["invalid"])["tasks"]
    assert not ecs.describe_tasks(cluster="dagster", tasks=["invalid"])["tasks"]

    ecs.register_task_definition(family="bridge", containerDefinitions=[], networkMode="bridge")

    default = ecs.run_task(taskDefinition="bridge")
    default_arn = default["tasks"][0]["taskArn"]
    default_id = default_arn.split("/")[-1]

    dagster = ecs.run_task(taskDefinition="bridge", cluster="dagster")
    dagster_arn = dagster["tasks"][0]["taskArn"]
    dagster_id = dagster_arn.split("/")[-1]

    # It uses the default cluster
    assert ecs.describe_tasks(tasks=[default_arn]) == default
    # It works with task ARNs
    assert not ecs.describe_tasks(tasks=[dagster_arn])["tasks"]
    assert ecs.describe_tasks(tasks=[dagster_arn], cluster="dagster") == dagster

    # And task IDs
    assert ecs.describe_tasks(tasks=[default_id]) == default
    assert not ecs.describe_tasks(tasks=[dagster_id])["tasks"]
    assert ecs.describe_tasks(tasks=[dagster_id], cluster="dagster") == dagster


def test_list_tags_for_resource(ecs):
    invalid_arn = ecs._task_arn("invalid")
    with pytest.raises(ClientError):
        # The task doesn't exist
        ecs.list_tags_for_resource(resourceArn=invalid_arn)

    tags = [{"key": "foo", "value": "bar"}, {"key": "fizz", "value": "buzz"}]
    ecs.register_task_definition(family="dagster", containerDefinitions=[], networkMode="bridge")
    arn = ecs.run_task(taskDefinition="dagster")["tasks"][0]["taskArn"]

    assert not ecs.list_tags_for_resource(resourceArn=arn)["tags"]

    ecs.tag_resource(resourceArn=arn, tags=tags)

    assert ecs.list_tags_for_resource(resourceArn=arn)["tags"] == tags


def test_list_task_definitions(ecs):
    assert not ecs.list_task_definitions()["taskDefinitionArns"]

    def arn(task_definition):
        return task_definition["taskDefinition"]["taskDefinitionArn"]

    dagster1 = arn(ecs.register_task_definition(family="dagster", containerDefinitions=[]))
    dagster2 = arn(ecs.register_task_definition(family="dagster", containerDefinitions=[]))
    other1 = arn(ecs.register_task_definition(family="other", containerDefinitions=[]))

    assert len(ecs.list_task_definitions()["taskDefinitionArns"]) == 3
    assert dagster1 in ecs.list_task_definitions()["taskDefinitionArns"]
    assert dagster2 in ecs.list_task_definitions()["taskDefinitionArns"]
    assert other1 in ecs.list_task_definitions()["taskDefinitionArns"]


def test_list_tasks(ecs):
    assert not ecs.list_tasks()["taskArns"]

    ecs.register_task_definition(family="dagster", containerDefinitions=[], networkMode="bridge")
    ecs.register_task_definition(family="other", containerDefinitions=[], networkMode="bridge")

    def arn(response):
        return response["tasks"][0]["taskArn"]

    default_cluster_dagster_family = arn(ecs.run_task(taskDefinition="dagster"))
    other_cluster_dagster_family = arn(ecs.run_task(taskDefinition="dagster", cluster="other"))
    default_cluster_other_family = arn(ecs.run_task(taskDefinition="other"))
    other_cluster_other_family = arn(ecs.run_task(taskDefinition="other", cluster="other"))

    # List using different combinations of cluster and family filters
    response = ecs.list_tasks()
    assert len(response["taskArns"]) == 2
    assert default_cluster_dagster_family in response["taskArns"]
    assert default_cluster_other_family in response["taskArns"]

    response = ecs.list_tasks(family="dagster")
    assert len(response["taskArns"]) == 1
    assert default_cluster_dagster_family in response["taskArns"]

    response = ecs.list_tasks(cluster="other")
    assert len(response["taskArns"]) == 2
    assert other_cluster_dagster_family in response["taskArns"]
    assert other_cluster_other_family in response["taskArns"]

    response = ecs.list_tasks(cluster="other", family="dagster")
    assert len(response["taskArns"]) == 1
    assert other_cluster_dagster_family in response["taskArns"]


def test_register_task_definition(ecs):
    response = ecs.register_task_definition(family="dagster", containerDefinitions=[])
    assert response["taskDefinition"]["family"] == "dagster"
    assert response["taskDefinition"]["revision"] == 1
    assert response["taskDefinition"]["taskDefinitionArn"].endswith("dagster:1")

    response = ecs.register_task_definition(family="other", containerDefinitions=[])
    assert response["taskDefinition"]["family"] == "other"
    assert response["taskDefinition"]["revision"] == 1
    assert response["taskDefinition"]["taskDefinitionArn"].endswith("other:1")

    response = ecs.register_task_definition(family="dagster", containerDefinitions=[])
    assert response["taskDefinition"]["family"] == "dagster"
    assert response["taskDefinition"]["revision"] == 2
    assert response["taskDefinition"]["taskDefinitionArn"].endswith("dagster:2")

    response = ecs.register_task_definition(
        family="dagster", containerDefinitions=[{"image": "hello_world:latest"}]
    )
    assert response["taskDefinition"]["containerDefinitions"][0]["image"] == "hello_world:latest"

    response = ecs.register_task_definition(
        family="dagster", containerDefinitions=[], networkMode="bridge"
    )
    assert response["taskDefinition"]["networkMode"] == "bridge"


def test_run_task(ecs, subnet):
    with pytest.raises(ParamValidationError):
        # The task doesn't exist
        ecs.run_task()

    with pytest.raises(ClientError):
        # The task definition doesn't exist
        ecs.run_task(taskDefinition="dagster")

    ecs.register_task_definition(family="awsvpc", containerDefinitions=[], networkMode="awsvpc")
    ecs.register_task_definition(family="bridge", containerDefinitions=[], networkMode="bridge")

    response = ecs.run_task(taskDefinition="bridge")
    assert len(response["tasks"]) == 1
    assert "bridge" in response["tasks"][0]["taskDefinitionArn"]
    assert response["tasks"][0]["lastStatus"] == "RUNNING"

    # It uses the default cluster
    assert response["tasks"][0]["clusterArn"] == ecs._cluster_arn("default")
    response = ecs.run_task(taskDefinition="bridge", cluster="dagster")
    assert response["tasks"][0]["clusterArn"] == ecs._cluster_arn("dagster")
    response = ecs.run_task(taskDefinition="bridge", cluster=ecs._cluster_arn("dagster"))
    assert response["tasks"][0]["clusterArn"] == ecs._cluster_arn("dagster")

    response = ecs.run_task(taskDefinition="bridge", count=2)
    assert len(response["tasks"]) == 2
    assert all(["bridge" in task["taskDefinitionArn"] for task in response["tasks"]])

    with pytest.raises(ClientError):
        # It must have a networkConfiguration if networkMode is "awsvpc"
        ecs.run_task(taskDefinition="awsvpc")

    # The subnet doesn't exist
    with pytest.raises(ClientError):
        ecs.run_task(
            taskDefinition="awsvpc",
            networkConfiguration={"awsvpcConfiguration": {"subnets": ["subnet-12345"]}},
        )

    #  The subnet exists but there's no network interface
    with pytest.raises(ClientError):
        ecs.run_task(
            taskDefinition="awsvpc",
            networkConfiguration={"awsvpcConfiguration": {"subnets": [subnet.id]}},
        )

    network_interface = subnet.create_network_interface()
    response = ecs.run_task(
        taskDefinition="awsvpc",
        networkConfiguration={"awsvpcConfiguration": {"subnets": [subnet.id]}},
    )

    assert len(response["tasks"]) == 1
    assert "awsvpc" in response["tasks"][0]["taskDefinitionArn"]
    attachment = response["tasks"][0]["attachments"][0]
    assert attachment["type"] == "ElasticNetworkInterface"
    assert {"name": "subnetId", "value": subnet.id} in attachment["details"]
    assert {"name": "networkInterfaceId", "value": network_interface.id} in attachment["details"]

    # containers and overrides are included
    ecs.register_task_definition(
        family="container",
        containerDefinitions=[
            {
                "name": "hello_world",
                "image": "hello_world:latest",
                "environment": [{"name": "FOO", "value": "bar"}],
            }
        ],
        networkMode="bridge",
    )
    response = ecs.run_task(taskDefinition="container")
    assert response["tasks"][0]["containers"]
    # ECS does not expose the task definition's environment when
    # describing tasks
    assert "FOO" not in response

    response = ecs.run_task(
        taskDefinition="container",
        overrides={"containerOverrides": [{"name": "hello_world", "command": ["ls"]}]},
    )
    assert response["tasks"][0]["overrides"]["containerOverrides"][0]["command"] == ["ls"]


def test_stop_task(ecs):
    with pytest.raises(ClientError):
        # The task doesn't exist
        ecs.stop_task(task=ecs._task_arn("invalid"))

    ecs.register_task_definition(family="bridge", containerDefinitions=[], networkMode="bridge")
    task_arn = ecs.run_task(taskDefinition="bridge")["tasks"][0]["taskArn"]

    assert ecs.describe_tasks(tasks=[task_arn])["tasks"][0]["lastStatus"] == "RUNNING"

    response = ecs.stop_task(task=task_arn)
    assert response["task"]["taskArn"] == task_arn
    assert response["task"]["lastStatus"] == "STOPPED"

    assert ecs.describe_tasks(tasks=[task_arn])["tasks"][0]["lastStatus"] == "STOPPED"


def test_tag_resource(ecs):
    tags = [{"key": "foo", "value": "bar"}]

    invalid_arn = ecs._task_arn("invalid")
    with pytest.raises(ClientError):
        # The task doesn't exist
        ecs.tag_resource(resourceArn=invalid_arn, tags=tags)

    ecs.register_task_definition(family="dagster", containerDefinitions=[], networkMode="bridge")
    arn = ecs.run_task(taskDefinition="dagster")["tasks"][0]["taskArn"]

    ecs.tag_resource(resourceArn=arn, tags=tags)



import copy
import gzip
import io
import threading
import time
from unittest import mock

import pytest
from dagster.utils.test import create_test_pipeline_execution_context
from dagster_aws.emr import EmrClusterState, EmrError, EmrJobRunner
from dagster_aws.utils.mrjob.utils import _boto3_now
from moto import mock_emr

REGION = "us-west-1"


@mock_emr
def test_emr_create_cluster(emr_cluster_config):
    context = create_test_pipeline_execution_context()
    cluster = EmrJobRunner(region=REGION)
    cluster_id = cluster.run_job_flow(context.log, emr_cluster_config)
    assert cluster_id.startswith("j-")


@mock_emr
def test_emr_add_tags_and_describe_cluster(emr_cluster_config):
    context = create_test_pipeline_execution_context()
    emr = EmrJobRunner(region=REGION)

    cluster_id = emr.run_job_flow(context.log, emr_cluster_config)

    emr.add_tags(context.log, {"foobar": "v1", "baz": "123"}, cluster_id)

    tags = emr.describe_cluster(cluster_id)["Cluster"]["Tags"]

    assert {"Key": "baz", "Value": "123"} in tags
    assert {"Key": "foobar", "Value": "v1"} in tags


@mock_emr
def test_emr_describe_cluster(emr_cluster_config):
    context = create_test_pipeline_execution_context()
    cluster = EmrJobRunner(region=REGION)
    cluster_id = cluster.run_job_flow(context.log, emr_cluster_config)
    cluster_info = cluster.describe_cluster(cluster_id)["Cluster"]
    assert cluster_info["Name"] == "test-emr"
    assert EmrClusterState(cluster_info["Status"]["State"]) == EmrClusterState.Waiting


@mock_emr
def test_emr_id_from_name(emr_cluster_config):
    context = create_test_pipeline_execution_context()
    cluster = EmrJobRunner(region=REGION)
    cluster_id = cluster.run_job_flow(context.log, emr_cluster_config)
    assert cluster.cluster_id_from_name("test-emr") == cluster_id

    with pytest.raises(EmrError) as exc_info:
        cluster.cluster_id_from_name("cluster-doesnt-exist")

    assert "cluster cluster-doesnt-exist not found in region us-west-1" in str(exc_info.value)


def test_emr_construct_step_dict():
    cmd = ["pip", "install", "dagster"]

    assert EmrJobRunner.construct_step_dict_for_command("test_step", cmd) == {
        "Name": "test_step",
        "ActionOnFailure": "CONTINUE",
        "HadoopJarStep": {"Jar": "command-runner.jar", "Args": cmd},
    }

    assert EmrJobRunner.construct_step_dict_for_command(
        "test_second_step", cmd, action_on_failure="CANCEL_AND_WAIT"
    ) == {
        "Name": "test_second_step",
        "ActionOnFailure": "CANCEL_AND_WAIT",
        "HadoopJarStep": {"Jar": "command-runner.jar", "Args": cmd},
    }


@mock_emr
def test_emr_log_location_for_cluster(emr_cluster_config, mock_s3_bucket):
    context = create_test_pipeline_execution_context()
    emr = EmrJobRunner(region=REGION)
    cluster_id = emr.run_job_flow(context.log, emr_cluster_config)
    assert emr.log_location_for_cluster(cluster_id) == (mock_s3_bucket.name, "elasticmapreduce/")

    # Should raise when the log URI is missing
    emr_cluster_config = copy.deepcopy(emr_cluster_config)
    del emr_cluster_config["LogUri"]
    cluster_id = emr.run_job_flow(context.log, emr_cluster_config)
    with pytest.raises(EmrError) as exc_info:
        emr.log_location_for_cluster(cluster_id)

    assert "Log URI not specified, cannot retrieve step execution logs" in str(exc_info.value)


@mock_emr
def test_emr_retrieve_logs(emr_cluster_config, mock_s3_bucket):
    context = create_test_pipeline_execution_context()
    emr = EmrJobRunner(region=REGION)
    cluster_id = emr.run_job_flow(context.log, emr_cluster_config)
    assert emr.log_location_for_cluster(cluster_id) == (mock_s3_bucket.name, "elasticmapreduce/")

    def create_log():
        time.sleep(0.5)
        out = io.BytesIO()
        with gzip.GzipFile(fileobj=out, mode="w") as fo:
            fo.write(b"some log")

        prefix = "elasticmapreduce/{cluster_id}/steps/{step_id}".format(
            cluster_id=cluster_id, step_id="s-123456123456"
        )

        for name in ["stdout.gz", "stderr.gz"]:
            mock_s3_bucket.Object(prefix + "/" + name).put(  # pylint: disable=no-member
                Body=out.getvalue()
            )

    thread = threading.Thread(target=create_log, args=())
    thread.daemon = True
    thread.start()

    stdout_log, stderr_log = emr.retrieve_logs_for_step_id(
        context.log, cluster_id, "s-123456123456"
    )
    assert stdout_log == "some log"
    assert stderr_log == "some log"


def test_wait_for_log(mock_s3_bucket):
    def create_log():
        time.sleep(0.5)
        out = io.BytesIO()
        with gzip.GzipFile(fileobj=out, mode="w") as fo:
            fo.write(b"foo bar")

        mock_s3_bucket.Object("some_log_file").put(Body=out.getvalue())  # pylint: disable=no-member

    thread = threading.Thread(target=create_log, args=())
    thread.daemon = True
    thread.start()

    context = create_test_pipeline_execution_context()
    emr = EmrJobRunner(region=REGION)
    res = emr.wait_for_log(
        context.log,
        log_bucket=mock_s3_bucket.name,
        log_key="some_log_file",
        waiter_delay=1,
        waiter_max_attempts=2,
    )
    assert res == "foo bar"

    with pytest.raises(EmrError) as exc_info:
        emr.wait_for_log(
            context.log,
            log_bucket=mock_s3_bucket.name,
            log_key="does_not_exist",
            waiter_delay=1,
            waiter_max_attempts=1,
        )
    assert "EMR log file did not appear on S3 after waiting" in str(exc_info.value)


@mock_emr
def test_is_emr_step_complete(emr_cluster_config):
    context = create_test_pipeline_execution_context()
    emr = EmrJobRunner(region=REGION, check_cluster_every=1)

    cluster_id = emr.run_job_flow(context.log, emr_cluster_config)

    step_name = "test_step"
    step_cmd = ["ls", "/"]
    step_ids = emr.add_job_flow_steps(
        context.log, cluster_id, [emr.construct_step_dict_for_command(step_name, step_cmd)]
    )

    def get_step_dict(step_id, step_state):
        return {
            "Step": {
                "Id": step_id,
                "Name": step_name,
                "Config": {"Jar": "command-runner.jar", "Properties": {}, "Args": step_cmd},
                "ActionOnFailure": "CONTINUE",
                "Status": {
                    "State": step_state,
                    "StateChangeReason": {"Message": "everything is hosed"},
                    "Timeline": {"StartDateTime": _boto3_now()},
                },
            },
        }

    emr_step_id = step_ids[0]
    describe_step_returns = [
        get_step_dict(emr_step_id, "PENDING"),
        get_step_dict(emr_step_id, "RUNNING"),
        get_step_dict(emr_step_id, "COMPLETED"),
        get_step_dict(emr_step_id, "FAILED"),
    ]
    with mock.patch.object(EmrJobRunner, "describe_step", side_effect=describe_step_returns):
        assert not emr.is_emr_step_complete(context.log, cluster_id, emr_step_id)
        assert not emr.is_emr_step_complete(context.log, cluster_id, emr_step_id)
        assert emr.is_emr_step_complete(context.log, cluster_id, emr_step_id)

        with pytest.raises(EmrError) as exc_info:
            emr.is_emr_step_complete(context.log, cluster_id, emr_step_id)
            assert "step failed" in str(exc_info.value)



from dagster import ModeDefinition, execute_pipeline, pipeline, solid
from dagster_aws.s3 import S3FileHandle, file_handle_to_s3, s3_file_manager, s3_resource


def create_file_handle_pipeline(temp_file_handle):
    @solid
    def emit_temp_handle(_):
        return temp_file_handle

    @pipeline(
        mode_defs=[
            ModeDefinition(resource_defs={"s3": s3_resource, "file_manager": s3_file_manager})
        ]
    )
    def test():
        return file_handle_to_s3(emit_temp_handle())

    return test


def test_successful_file_handle_to_s3(mock_s3_bucket):
    foo_bytes = b"foo"
    remote_s3_object = mock_s3_bucket.Object("some-key/foo")
    remote_s3_object.put(Body=foo_bytes)

    file_handle = S3FileHandle(mock_s3_bucket.name, "some-key/foo")
    result = execute_pipeline(
        create_file_handle_pipeline(file_handle),
        run_config={
            "solids": {
                "file_handle_to_s3": {"config": {"Bucket": mock_s3_bucket.name, "Key": "some-key"}}
            },
            "resources": {"file_manager": {"config": {"s3_bucket": mock_s3_bucket.name}}},
        },
    )

    assert result.success

    assert mock_s3_bucket.Object(key="some-key").get()["Body"].read() == foo_bytes

    materializations = result.result_for_solid("file_handle_to_s3").materializations_during_compute
    assert len(materializations) == 1
    assert len(materializations[0].metadata_entries) == 1
    assert materializations[0].metadata_entries[
        0
    ].entry_data.path == "s3://{bucket}/some-key".format(bucket=mock_s3_bucket.name)
    assert materializations[0].metadata_entries[0].label == "some-key"
